#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or https://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
# --- Auto-generated code: don't edit this file ---
#
UNICODE_BLOCKS_VER_2_0_0 = {
    'Basic Latin': '\u0000-\u007F',
    'Latin-1 Supplement': '\u0080-\u00FF',
    'Latin Extended-A': '\u0100-\u017F',
    'Latin Extended-B': '\u0180-\u024F',
    'IPA Extensions': '\u0250-\u02AF',
    'Spacing Modifier Letters': '\u02B0-\u02FF',
    'Combining Diacritical Marks': '\u0300-\u036F',
    'Greek': '\u0370-\u03FF',
    'Cyrillic': '\u0400-\u04FF',
    'Armenian': '\u0530-\u058F',
    'Hebrew': '\u0590-\u05FF',
    'Arabic': '\u0600-\u06FF',
    'Devanagari': '\u0900-\u097F',
    'Bengali': '\u0980-\u09FF',
    'Gurmukhi': '\u0A00-\u0A7F',
    'Gujarati': '\u0A80-\u0AFF',
    'Oriya': '\u0B00-\u0B7F',
    'Tamil': '\u0B80-\u0BFF',
    'Telugu': '\u0C00-\u0C7F',
    'Kannada': '\u0C80-\u0CFF',
    'Malayalam': '\u0D00-\u0D7F',
    'Thai': '\u0E00-\u0E7F',
    'Lao': '\u0E80-\u0EFF',
    'Tibetan': '\u0F00-\u0FBF',
    'Georgian': '\u10A0-\u10FF',
    'Hangul Jamo': '\u1100-\u11FF',
    'Latin Extended Additional': '\u1E00-\u1EFF',
    'Greek Extended': '\u1F00-\u1FFF',
    'General Punctuation': '\u2000-\u206F',
    'Superscripts and Subscripts': '\u2070-\u209F',
    'Currency Symbols': '\u20A0-\u20CF',
    'Combining Marks for Symbols': '\u20D0-\u20FF',
    'Letterlike Symbols': '\u2100-\u214F',
    'Number Forms': '\u2150-\u218F',
    'Arrows': '\u2190-\u21FF',
    'Mathematical Operators': '\u2200-\u22FF',
    'Miscellaneous Technical': '\u2300-\u23FF',
    'Control Pictures': '\u2400-\u243F',
    'Optical Character Recognition': '\u2440-\u245F',
    'Enclosed Alphanumerics': '\u2460-\u24FF',
    'Box Drawing': '\u2500-\u257F',
    'Block Elements': '\u2580-\u259F',
    'Geometric Shapes': '\u25A0-\u25FF',
    'Miscellaneous Symbols': '\u2600-\u26FF',
    'Dingbats': '\u2700-\u27BF',
    'CJK Symbols and Punctuation': '\u3000-\u303F',
    'Hiragana': '\u3040-\u309F',
    'Katakana': '\u30A0-\u30FF',
    'Bopomofo': '\u3100-\u312F',
    'Hangul Compatibility Jamo': '\u3130-\u318F',
    'Kanbun': '\u3190-\u319F',
    'Enclosed CJK Letters and Months': '\u3200-\u32FF',
    'CJK Compatibility': '\u3300-\u33FF',
    'CJK Unified Ideographs': '\u4E00-\u9FFF',
    'Hangul Syllables': '\uAC00-\uD7A3',
    'High Surrogates': '\uD800-\uDB7F',
    'High Private Use Surrogates': '\uDB80-\uDBFF',
    'Low Surrogates': '\uDC00-\uDFFF',
    'Private Use': '\uE000-\uF8FF',
    'CJK Compatibility Ideographs': '\uF900-\uFAFF',
    'Alphabetic Presentation Forms': '\uFB00-\uFB4F',
    'Arabic Presentation Forms-A': '\uFB50-\uFDFF',
    'Combining Half Marks': '\uFE20-\uFE2F',
    'CJK Compatibility Forms': '\uFE30-\uFE4F',
    'Small Form Variants': '\uFE50-\uFE6F',
    'Arabic Presentation Forms-B': '\uFE70-\uFEFF',
    'Halfwidth and Fullwidth Forms': '\uFF00-\uFFEF',
    'Specials': '\uFEFF-\uFEFF\uFFF0-\uFFFF'
}

UPDATE_BLOCKS_VER_2_1_9 = {
    'Arabic Presentation Forms-B': '\uFE70-\uFEFE',
    'Specials': '\uFEFF-\uFEFF\uFFF0-\uFFFD'
}

UPDATE_BLOCKS_VER_3_0_0 = {
    'Syriac': '\u0700-\u074F',
    'Thaana': '\u0780-\u07BF',
    'Sinhala': '\u0D80-\u0DFF',
    'Tibetan': '\u0F00-\u0FFF',
    'Myanmar': '\u1000-\u109F',
    'Ethiopic': '\u1200-\u137F',
    'Cherokee': '\u13A0-\u13FF',
    'Unified Canadian Aboriginal Syllabics': '\u1400-\u167F',
    'Ogham': '\u1680-\u169F',
    'Runic': '\u16A0-\u16FF',
    'Khmer': '\u1780-\u17FF',
    'Mongolian': '\u1800-\u18AF',
    'Braille Patterns': '\u2800-\u28FF',
    'CJK Radicals Supplement': '\u2E80-\u2EFF',
    'Kangxi Radicals': '\u2F00-\u2FDF',
    'Ideographic Description Characters': '\u2FF0-\u2FFF',
    'Bopomofo Extended': '\u31A0-\u31BF',
    'CJK Unified Ideographs Extension A': '\u3400-\u4DB5',
    'Yi Syllables': '\uA000-\uA48F',
    'Yi Radicals': '\uA490-\uA4CF'
}

UPDATE_BLOCKS_VER_3_1_0 = {
    'Private Use': '\uE000-\uF8FF\U000F0000-\U000FFFFD\U00100000-\U0010FFFD',
    'Old Italic': '\U00010300-\U0001032F',
    'Gothic': '\U00010330-\U0001034F',
    'Deseret': '\U00010400-\U0001044F',
    'Byzantine Musical Symbols': '\U0001D000-\U0001D0FF',
    'Musical Symbols': '\U0001D100-\U0001D1FF',
    'Mathematical Alphanumeric Symbols': '\U0001D400-\U0001D7FF',
    'CJK Unified Ideographs Extension B': '\U00020000-\U0002A6D6',
    'CJK Compatibility Ideographs Supplement': '\U0002F800-\U0002FA1F',
    'Tags': '\U000E0000-\U000E007F'
}

REMOVED_BLOCKS_VER_3_1_1 = [
    'Greek', 'Combining Marks for Symbols', 'Private Use'
]

UPDATE_BLOCKS_VER_3_1_1 = {
    'Greek and Coptic': '\u0370-\u03FF',
    'Cyrillic Supplementary': '\u0500-\u052F',
    'Tagalog': '\u1700-\u171F',
    'Hanunoo': '\u1720-\u173F',
    'Buhid': '\u1740-\u175F',
    'Tagbanwa': '\u1760-\u177F',
    'Combining Diacritical Marks for Symbols': '\u20D0-\u20FF',
    'Miscellaneous Mathematical Symbols-A': '\u27C0-\u27EF',
    'Supplemental Arrows-A': '\u27F0-\u27FF',
    'Supplemental Arrows-B': '\u2900-\u297F',
    'Miscellaneous Mathematical Symbols-B': '\u2980-\u29FF',
    'Supplemental Mathematical Operators': '\u2A00-\u2AFF',
    'Katakana Phonetic Extensions': '\u31F0-\u31FF',
    'CJK Unified Ideographs Extension A': '\u3400-\u4DBF',
    'Hangul Syllables': '\uAC00-\uD7AF',
    'Private Use Area': '\uE000-\uF8FF',
    'Variation Selectors': '\uFE00-\uFE0F',
    'Arabic Presentation Forms-B': '\uFE70-\uFEFF',
    'Specials': '\uFFF0-\uFFFF',
    'CJK Unified Ideographs Extension B': '\U00020000-\U0002A6DF',
    'Supplementary Private Use Area-A': '\U000F0000-\U000FFFFF',
    'Supplementary Private Use Area-B': '\U00100000-\U0010FFFF'
}

UPDATE_BLOCKS_VER_4_0_0 = {
    'Limbu': '\u1900-\u194F',
    'Tai Le': '\u1950-\u197F',
    'Khmer Symbols': '\u19E0-\u19FF',
    'Phonetic Extensions': '\u1D00-\u1D7F',
    'Miscellaneous Symbols and Arrows': '\u2B00-\u2BFF',
    'Yijing Hexagram Symbols': '\u4DC0-\u4DFF',
    'Linear B Syllabary': '\U00010000-\U0001007F',
    'Linear B Ideograms': '\U00010080-\U000100FF',
    'Aegean Numbers': '\U00010100-\U0001013F',
    'Ugaritic': '\U00010380-\U0001039F',
    'Shavian': '\U00010450-\U0001047F',
    'Osmanya': '\U00010480-\U000104AF',
    'Cypriot Syllabary': '\U00010800-\U0001083F',
    'Tai Xuan Jing Symbols': '\U0001D300-\U0001D35F',
    'Variation Selectors Supplement': '\U000E0100-\U000E01EF'
}

REMOVED_BLOCKS_VER_4_0_1 = [
    'Cyrillic Supplementary'
]

UPDATE_BLOCKS_VER_4_0_1 = {
    'Cyrillic Supplement': '\u0500-\u052F'
}

UPDATE_BLOCKS_VER_4_1_0 = {
    'Arabic Supplement': '\u0750-\u077F',
    'Ethiopic Supplement': '\u1380-\u139F',
    'New Tai Lue': '\u1980-\u19DF',
    'Buginese': '\u1A00-\u1A1F',
    'Phonetic Extensions Supplement': '\u1D80-\u1DBF',
    'Combining Diacritical Marks Supplement': '\u1DC0-\u1DFF',
    'Glagolitic': '\u2C00-\u2C5F',
    'Coptic': '\u2C80-\u2CFF',
    'Georgian Supplement': '\u2D00-\u2D2F',
    'Tifinagh': '\u2D30-\u2D7F',
    'Ethiopic Extended': '\u2D80-\u2DDF',
    'Supplemental Punctuation': '\u2E00-\u2E7F',
    'CJK Strokes': '\u31C0-\u31EF',
    'Modifier Tone Letters': '\uA700-\uA71F',
    'Syloti Nagri': '\uA800-\uA82F',
    'Vertical Forms': '\uFE10-\uFE1F',
    'Ancient Greek Numbers': '\U00010140-\U0001018F',
    'Old Persian': '\U000103A0-\U000103DF',
    'Kharoshthi': '\U00010A00-\U00010A5F',
    'Ancient Greek Musical Notation': '\U0001D200-\U0001D24F'
}

UPDATE_BLOCKS_VER_5_0_0 = {
    'NKo': '\u07C0-\u07FF',
    'Balinese': '\u1B00-\u1B7F',
    'Latin Extended-C': '\u2C60-\u2C7F',
    'Latin Extended-D': '\uA720-\uA7FF',
    'Phags-pa': '\uA840-\uA87F',
    'Phoenician': '\U00010900-\U0001091F',
    'Cuneiform': '\U00012000-\U000123FF',
    'Cuneiform Numbers and Punctuation': '\U00012400-\U0001247F',
    'Counting Rod Numerals': '\U0001D360-\U0001D37F'
}

UPDATE_BLOCKS_VER_5_1_0 = {
    'Sundanese': '\u1B80-\u1BBF',
    'Lepcha': '\u1C00-\u1C4F',
    'Ol Chiki': '\u1C50-\u1C7F',
    'Cyrillic Extended-A': '\u2DE0-\u2DFF',
    'Vai': '\uA500-\uA63F',
    'Cyrillic Extended-B': '\uA640-\uA69F',
    'Saurashtra': '\uA880-\uA8DF',
    'Kayah Li': '\uA900-\uA92F',
    'Rejang': '\uA930-\uA95F',
    'Cham': '\uAA00-\uAA5F',
    'Ancient Symbols': '\U00010190-\U000101CF',
    'Phaistos Disc': '\U000101D0-\U000101FF',
    'Lycian': '\U00010280-\U0001029F',
    'Carian': '\U000102A0-\U000102DF',
    'Lydian': '\U00010920-\U0001093F',
    'Mahjong Tiles': '\U0001F000-\U0001F02F',
    'Domino Tiles': '\U0001F030-\U0001F09F'
}

UPDATE_BLOCKS_VER_5_2_0 = {
    'Samaritan': '\u0800-\u083F',
    'Unified Canadian Aboriginal Syllabics Extended': '\u18B0-\u18FF',
    'Tai Tham': '\u1A20-\u1AAF',
    'Vedic Extensions': '\u1CD0-\u1CFF',
    'Lisu': '\uA4D0-\uA4FF',
    'Bamum': '\uA6A0-\uA6FF',
    'Common Indic Number Forms': '\uA830-\uA83F',
    'Devanagari Extended': '\uA8E0-\uA8FF',
    'Hangul Jamo Extended-A': '\uA960-\uA97F',
    'Javanese': '\uA980-\uA9DF',
    'Myanmar Extended-A': '\uAA60-\uAA7F',
    'Tai Viet': '\uAA80-\uAADF',
    'Meetei Mayek': '\uABC0-\uABFF',
    'Hangul Jamo Extended-B': '\uD7B0-\uD7FF',
    'Imperial Aramaic': '\U00010840-\U0001085F',
    'Old South Arabian': '\U00010A60-\U00010A7F',
    'Avestan': '\U00010B00-\U00010B3F',
    'Inscriptional Parthian': '\U00010B40-\U00010B5F',
    'Inscriptional Pahlavi': '\U00010B60-\U00010B7F',
    'Old Turkic': '\U00010C00-\U00010C4F',
    'Rumi Numeral Symbols': '\U00010E60-\U00010E7F',
    'Kaithi': '\U00011080-\U000110CF',
    'Egyptian Hieroglyphs': '\U00013000-\U0001342F',
    'Enclosed Alphanumeric Supplement': '\U0001F100-\U0001F1FF',
    'Enclosed Ideographic Supplement': '\U0001F200-\U0001F2FF',
    'CJK Unified Ideographs Extension C': '\U0002A700-\U0002B73F'
}

UPDATE_BLOCKS_VER_6_0_0 = {
    'Mandaic': '\u0840-\u085F',
    'Batak': '\u1BC0-\u1BFF',
    'Ethiopic Extended-A': '\uAB00-\uAB2F',
    'Brahmi': '\U00011000-\U0001107F',
    'Bamum Supplement': '\U00016800-\U00016A3F',
    'Kana Supplement': '\U0001B000-\U0001B0FF',
    'Playing Cards': '\U0001F0A0-\U0001F0FF',
    'Miscellaneous Symbols And Pictographs': '\U0001F300-\U0001F5FF',
    'Emoticons': '\U0001F600-\U0001F64F',
    'Transport And Map Symbols': '\U0001F680-\U0001F6FF',
    'Alchemical Symbols': '\U0001F700-\U0001F77F',
    'CJK Unified Ideographs Extension D': '\U0002B740-\U0002B81F'
}

UPDATE_BLOCKS_VER_6_1_0 = {
    'Arabic Extended-A': '\u08A0-\u08FF',
    'Sundanese Supplement': '\u1CC0-\u1CCF',
    'Meetei Mayek Extensions': '\uAAE0-\uAAFF',
    'Meroitic Hieroglyphs': '\U00010980-\U0001099F',
    'Meroitic Cursive': '\U000109A0-\U000109FF',
    'Sora Sompeng': '\U000110D0-\U000110FF',
    'Chakma': '\U00011100-\U0001114F',
    'Sharada': '\U00011180-\U000111DF',
    'Takri': '\U00011680-\U000116CF',
    'Miao': '\U00016F00-\U00016F9F',
    'Arabic Mathematical Alphabetic Symbols': '\U0001EE00-\U0001EEFF'
}

REMOVED_BLOCKS_VER_7_0_0 = [
    'Miscellaneous Symbols And Pictographs', 'Transport And Map Symbols'
]

UPDATE_BLOCKS_VER_7_0_0 = {
    'Combining Diacritical Marks Extended': '\u1AB0-\u1AFF',
    'Myanmar Extended-B': '\uA9E0-\uA9FF',
    'Latin Extended-E': '\uAB30-\uAB6F',
    'Coptic Epact Numbers': '\U000102E0-\U000102FF',
    'Old Permic': '\U00010350-\U0001037F',
    'Elbasan': '\U00010500-\U0001052F',
    'Caucasian Albanian': '\U00010530-\U0001056F',
    'Linear A': '\U00010600-\U0001077F',
    'Palmyrene': '\U00010860-\U0001087F',
    'Nabataean': '\U00010880-\U000108AF',
    'Old North Arabian': '\U00010A80-\U00010A9F',
    'Manichaean': '\U00010AC0-\U00010AFF',
    'Psalter Pahlavi': '\U00010B80-\U00010BAF',
    'Mahajani': '\U00011150-\U0001117F',
    'Sinhala Archaic Numbers': '\U000111E0-\U000111FF',
    'Khojki': '\U00011200-\U0001124F',
    'Khudawadi': '\U000112B0-\U000112FF',
    'Grantha': '\U00011300-\U0001137F',
    'Tirhuta': '\U00011480-\U000114DF',
    'Siddham': '\U00011580-\U000115FF',
    'Modi': '\U00011600-\U0001165F',
    'Warang Citi': '\U000118A0-\U000118FF',
    'Pau Cin Hau': '\U00011AC0-\U00011AFF',
    'Mro': '\U00016A40-\U00016A6F',
    'Bassa Vah': '\U00016AD0-\U00016AFF',
    'Pahawh Hmong': '\U00016B00-\U00016B8F',
    'Duployan': '\U0001BC00-\U0001BC9F',
    'Shorthand Format Controls': '\U0001BCA0-\U0001BCAF',
    'Mende Kikakui': '\U0001E800-\U0001E8DF',
    'Miscellaneous Symbols and Pictographs': '\U0001F300-\U0001F5FF',
    'Ornamental Dingbats': '\U0001F650-\U0001F67F',
    'Transport and Map Symbols': '\U0001F680-\U0001F6FF',
    'Geometric Shapes Extended': '\U0001F780-\U0001F7FF',
    'Supplemental Arrows-C': '\U0001F800-\U0001F8FF'
}

UPDATE_BLOCKS_VER_8_0_0 = {
    'Cherokee Supplement': '\uAB70-\uABBF',
    'Hatran': '\U000108E0-\U000108FF',
    'Old Hungarian': '\U00010C80-\U00010CFF',
    'Multani': '\U00011280-\U000112AF',
    'Ahom': '\U00011700-\U0001173F',
    'Early Dynastic Cuneiform': '\U00012480-\U0001254F',
    'Anatolian Hieroglyphs': '\U00014400-\U0001467F',
    'Sutton SignWriting': '\U0001D800-\U0001DAAF',
    'Supplemental Symbols and Pictographs': '\U0001F900-\U0001F9FF',
    'CJK Unified Ideographs Extension E': '\U0002B820-\U0002CEAF'
}

UPDATE_BLOCKS_VER_9_0_0 = {
    'Cyrillic Extended-C': '\u1C80-\u1C8F',
    'Osage': '\U000104B0-\U000104FF',
    'Newa': '\U00011400-\U0001147F',
    'Mongolian Supplement': '\U00011660-\U0001167F',
    'Bhaiksuki': '\U00011C00-\U00011C6F',
    'Marchen': '\U00011C70-\U00011CBF',
    'Ideographic Symbols and Punctuation': '\U00016FE0-\U00016FFF',
    'Tangut': '\U00017000-\U000187FF',
    'Tangut Components': '\U00018800-\U00018AFF',
    'Glagolitic Supplement': '\U0001E000-\U0001E02F',
    'Adlam': '\U0001E900-\U0001E95F'
}

UPDATE_BLOCKS_VER_10_0_0 = {
    'Syriac Supplement': '\u0860-\u086F',
    'Zanabazar Square': '\U00011A00-\U00011A4F',
    'Soyombo': '\U00011A50-\U00011AAF',
    'Masaram Gondi': '\U00011D00-\U00011D5F',
    'Kana Extended-A': '\U0001B100-\U0001B12F',
    'Nushu': '\U0001B170-\U0001B2FF',
    'CJK Unified Ideographs Extension F': '\U0002CEB0-\U0002EBEF'
}

UPDATE_BLOCKS_VER_11_0_0 = {
    'Georgian Extended': '\u1C90-\u1CBF',
    'Hanifi Rohingya': '\U00010D00-\U00010D3F',
    'Old Sogdian': '\U00010F00-\U00010F2F',
    'Sogdian': '\U00010F30-\U00010F6F',
    'Dogra': '\U00011800-\U0001184F',
    'Gunjala Gondi': '\U00011D60-\U00011DAF',
    'Makasar': '\U00011EE0-\U00011EFF',
    'Medefaidrin': '\U00016E40-\U00016E9F',
    'Mayan Numerals': '\U0001D2E0-\U0001D2FF',
    'Indic Siyaq Numbers': '\U0001EC70-\U0001ECBF',
    'Chess Symbols': '\U0001FA00-\U0001FA6F'
}

UPDATE_BLOCKS_VER_12_0_0 = {
    'Elymaic': '\U00010FE0-\U00010FFF',
    'Nandinagari': '\U000119A0-\U000119FF',
    'Tamil Supplement': '\U00011FC0-\U00011FFF',
    'Egyptian Hieroglyph Format Controls': '\U00013430-\U0001343F',
    'Small Kana Extension': '\U0001B130-\U0001B16F',
    'Nyiakeng Puachue Hmong': '\U0001E100-\U0001E14F',
    'Wancho': '\U0001E2C0-\U0001E2FF',
    'Ottoman Siyaq Numbers': '\U0001ED00-\U0001ED4F',
    'Symbols and Pictographs Extended-A': '\U0001FA70-\U0001FAFF'
}

UPDATE_BLOCKS_VER_13_0_0 = {
    'Yezidi': '\U00010E80-\U00010EBF',
    'Chorasmian': '\U00010FB0-\U00010FDF',
    'Dives Akuru': '\U00011900-\U0001195F',
    'Lisu Supplement': '\U00011FB0-\U00011FBF',
    'Khitan Small Script': '\U00018B00-\U00018CFF',
    'Tangut Supplement': '\U00018D00-\U00018D8F',
    'Symbols for Legacy Computing': '\U0001FB00-\U0001FBFF',
    'CJK Unified Ideographs Extension G': '\U00030000-\U0003134F'
}

UPDATE_BLOCKS_VER_14_0_0 = {
    'Arabic Extended-B': '\u0870-\u089F',
    'Vithkuqi': '\U00010570-\U000105BF',
    'Latin Extended-F': '\U00010780-\U000107BF',
    'Old Uyghur': '\U00010F70-\U00010FAF',
    'Ahom': '\U00011700-\U0001174F',
    'Unified Canadian Aboriginal Syllabics Extended-A': '\U00011AB0-\U00011ABF',
    'Cypro-Minoan': '\U00012F90-\U00012FFF',
    'Tangsa': '\U00016A70-\U00016ACF',
    'Tangut Supplement': '\U00018D00-\U00018D7F',
    'Kana Extended-B': '\U0001AFF0-\U0001AFFF',
    'Znamenny Musical Notation': '\U0001CF00-\U0001CFCF',
    'Latin Extended-G': '\U0001DF00-\U0001DFFF',
    'Toto': '\U0001E290-\U0001E2BF',
    'Ethiopic Extended-B': '\U0001E7E0-\U0001E7FF'
}

UPDATE_BLOCKS_VER_15_0_0 = {
    'Arabic Extended-C': '\U00010EC0-\U00010EFF',
    'Devanagari Extended-A': '\U00011B00-\U00011B5F',
    'Kawi': '\U00011F00-\U00011F5F',
    'Egyptian Hieroglyph Format Controls': '\U00013430-\U0001345F',
    'Kaktovik Numerals': '\U0001D2C0-\U0001D2DF',
    'Cyrillic Extended-D': '\U0001E030-\U0001E08F',
    'Nag Mundari': '\U0001E4D0-\U0001E4FF',
    'CJK Unified Ideographs Extension H': '\U00031350-\U000323AF'
}

UPDATE_BLOCKS_VER_15_1_0 = {
    'CJK Unified Ideographs Extension I': '\U0002EBF0-\U0002EE5F'
}

UPDATE_BLOCKS_VER_16_0_0 = {
    'Todhri': '\U000105C0-\U000105FF',
    'Garay': '\U00010D40-\U00010D8F',
    'Tulu-Tigalari': '\U00011380-\U000113FF',
    'Myanmar Extended-C': '\U000116D0-\U000116FF',
    'Sunuwar': '\U00011BC0-\U00011BFF',
    'Egyptian Hieroglyphs Extended-A': '\U00013460-\U000143FF',
    'Gurung Khema': '\U00016100-\U0001613F',
    'Kirat Rai': '\U00016D40-\U00016D7F',
    'Symbols for Legacy Computing Supplement': '\U0001CC00-\U0001CEBF',
    'Ol Onal': '\U0001E5D0-\U0001E5FF'
}

UPDATE_BLOCKS_VER_17_0_0 = {
    'Sidetic': '\U00010940-\U0001095F',
    'Sharada Supplement': '\U00011B60-\U00011B7F',
    'Tolong Siki': '\U00011DB0-\U00011DEF',
    'Beria Erfe': '\U00016EA0-\U00016EDF',
    'Tangut Components Supplement': '\U00018D80-\U00018DFF',
    'Miscellaneous Symbols Supplement': '\U0001CEC0-\U0001CEFF',
    'Tai Yo': '\U0001E6C0-\U0001E6FF',
    'CJK Unified Ideographs Extension J': '\U000323B0-\U0003347F'
}
